import sys
from typing import Tuple

import aiohttp
import pytest

from coiled.exceptions import (
    AccountFormatError,
    CidrInvalidError,
    GPUTypeError,
    ParseIdentifierError,
    PortValidationError,
)
from coiled.utils import (
    bytes_to_mb,
    get_account_membership,
    get_instance_type_from_cpu_memory,
    get_platform,
    handle_credentials,
    has_program_quota,
    is_gcp,
    normalize_server,
    parse_backend_options,
    parse_gcp_region_zone,
    parse_identifier,
    parse_requested_memory,
    parse_wait_for_workers,
    validate_account,
    validate_cidr_block,
    validate_gpu_type,
    validate_ports,
)


@pytest.mark.parametrize(
    "identifier,expected",
    [
        ("coiled/xgboost:1efd34", ("coiled", "xgboost", "1efd34")),
        ("xgboost:1efd34", (None, "xgboost", "1efd34")),
        ("coiled/xgboost", ("coiled", "xgboost", None)),
        ("xgboost", (None, "xgboost", None)),
        ("coiled/xgboost-py37", ("coiled", "xgboost-py37", None)),
        ("xgboost_py38", (None, "xgboost_py38", None)),
    ],
)
def test_parse_good_names(identifier, expected: Tuple[str, str]):
    account, name, revision = parse_identifier(identifier, property_name="name_that_would_be_printed_in_error")
    assert (account, name, revision) == expected


@pytest.mark.parametrize(
    "identifier",
    [
        "coiled/dan/xgboost",
        "coiled/dan?xgboost",
        "dan\\xgboost",
        "jimmy/xgbóst",
        "My-Test-Env",
        "My Test Env",
        "coiled/My-Test-Env:abc123",
        "coiled/my-test-env!",
        "Coiled/my-test-env",
        "",
    ],
)
def test_parse_bad_names(identifier):
    with pytest.raises(ParseIdentifierError, match="software_environment"):
        parse_identifier(identifier, property_name="software_environment")


def test_parse_cluster_uppercase_names():
    cluster_name = "BobTheBuilder-1j850d-b"
    account, name, revision = parse_identifier(cluster_name, allow_uppercase=True)

    assert name == cluster_name
    assert account is None
    assert revision is None


def test_get_platform(monkeypatch):
    with monkeypatch.context() as m:
        monkeypatch.setattr(sys, "platform", "linux")
        assert get_platform() == "linux"

    with monkeypatch.context() as m:
        m.setattr(sys, "platform", "darwin")
        assert get_platform() == "osx"

    with monkeypatch.context() as m:
        m.setattr(sys, "platform", "win32")
        assert get_platform() == "windows"

    with monkeypatch.context() as m:
        m.setattr(sys, "platform", "bad-platform")
        with pytest.raises(ValueError) as result:
            assert get_platform() == "windows"

        err_msg = str(result).lower()
        assert "invalid" in err_msg
        assert "bad-platform" in err_msg


def test_normalize_server():
    assert normalize_server("http://beta.coiledhq.com") == "https://cloud.coiled.io"
    assert normalize_server("https://beta.coiled.io") == "https://cloud.coiled.io"


def test_get_account_membership():
    assert get_account_membership({}, account=None) is None

    response = {"membership_set": []}
    assert get_account_membership(response, account=None) is None

    membership = {
        "account": {"slug": "coiled"},
    }
    response = {"membership_set": [membership], "username": "not-coiled", "default_account": {"slug": "coiled"}}
    assert get_account_membership(response, account="coiled") == membership
    assert get_account_membership(response, account="test_user") != membership

    response["username"] = "coiled"
    assert get_account_membership(response) == membership


def test_get_account_membership_uppercase_account():
    assert get_account_membership({}, account=None) is None

    response = {"membership_set": []}
    assert get_account_membership(response, account=None) is None

    membership = {
        "account": {"slug": "coiled"},
    }
    response = {"membership_set": [membership], "username": "not-coiled", "default_account": {"slug": "coiled"}}
    assert get_account_membership(response, account="coiled") == membership
    assert get_account_membership(response, account="test_user") != membership

    response["username"] = "Coiled"
    assert get_account_membership(response) == membership


def test_has_program_quota():
    assert has_program_quota({}) is False

    membership = {"account": {"active_program": {"has_quota": True}}}
    assert has_program_quota(membership) is True


def test_gcp_region_zone():
    region, zone = parse_gcp_region_zone(region="us-central1")

    assert region == "us-central1"
    assert zone is None

    region, zone = parse_gcp_region_zone(zone="us-east1-c")

    assert region == "us-east1"
    assert zone == "us-east1-c"

    region, zone = parse_gcp_region_zone(region="test-region", zone="test-zone")

    assert region == "test-region"
    assert zone == "test-zone"

    region, zone = parse_gcp_region_zone(region="us-central1", zone="b")

    assert region == "us-central1"
    assert zone == "us-central1-b"

    region, zone = parse_gcp_region_zone(zone="a")

    assert region == "us-east1"
    assert zone == "us-east1-a"


def test_validate_gpu_type():
    is_valid = validate_gpu_type("nvidia-tesla-t4")
    assert is_valid

    with pytest.raises(GPUTypeError) as e:
        validate_gpu_type("T100")

    assert "GPU type 'T100' is not a supported GPU type" in e.value.args[0]


@pytest.mark.parametrize(
    "account,backend,provider_name,expected",
    [
        ("alice", "vm_gcp", "gcp", True),
        ("alice", "vm", "gcp", True),
        ("alice", "vm_aws", "aws", False),
        ("alice", "vm", "aws", False),
    ],
)
def test_is_gcp(account, backend, provider_name, expected):
    accounts = {"alice": {"backend": backend, "options": {"provider_name": provider_name}}}
    assert is_gcp(account, accounts) == expected


@pytest.mark.parametrize(
    "backend_options,account,backend,provider_name,worker_gpu,expected",
    [
        ({"preemptible": False}, "alice", "vm_gcp", "gcp", None, {"spot": False}),
        ({"preemptible": False}, "alice", "vm", "aws", None, {"preemptible": False}),
        ({"preemptible": False}, "bob", "vm", "gcp", None, {"spot": False}),
        ({}, "bob", "vm", "gcp", 0, {}),
        ({}, "bob", "vm", "gcp", 1, {"spot": False}),
    ],
)
def test_parse_backend_options(backend_options, account, backend, provider_name, worker_gpu, expected):
    accounts = {"alice": {"backend": backend}, "bob": {"backend": backend}}
    if provider_name is not None:
        accounts["alice"]["options"] = {"provider_name": provider_name}
        accounts["bob"]["options"] = {"provider_name": provider_name}
    parsed_backend_options = parse_backend_options(backend_options, account, accounts, worker_gpu)
    assert parsed_backend_options == expected


@pytest.mark.parametrize(
    "backend_options,account,backend,provider_name,worker_gpu,expected",
    [
        (
            {"firewall": {"ports": [12, 34, 59]}},
            "alice",
            "vm_gcp",
            "gcp",
            None,
            {"firewall": {"ports": [12, 34, 59]}},
        ),
        (
            {"firewall": {"cidr": "0.1.2.3/45"}},
            "bobthebuilder",
            "vm",
            "aws",
            None,
            {"firewall": {"cidr": "0.1.2.3/45"}},
        ),
    ],
)
def test_parse_backend_options_firewall(backend_options, account, backend, provider_name, worker_gpu, expected):
    accounts = {
        "alice": {"backend": backend},
        "hi!": {"backend": backend},
        "bobthebuilder": {"backend": backend},
    }
    parsed_backend_options = parse_backend_options(backend_options, account, accounts, worker_gpu)
    assert parsed_backend_options == expected


@pytest.mark.parametrize(
    "backend_options,account,backend,provider_name,worker_gpu",
    [
        (
            {"firewall": {"ports": "12,45,67"}},
            "alice",
            "vm_gcp",
            "gcp",
            None,
        ),
        (
            {"firewall": {"ports": [12, 345, "hi"]}},
            "bobthebuilder",
            "vm",
            "aws",
            None,
        ),
    ],
)
def test_parse_backend_options_firewall_port_validation_raises(
    backend_options, account, backend, provider_name, worker_gpu
):
    accounts = {
        "alice": {"backend": backend},
        "hi!": {"backend": backend},
        "bobthebuilder": {"backend": backend},
    }
    with pytest.raises(PortValidationError):
        parse_backend_options(backend_options, account, accounts, worker_gpu)


@pytest.mark.parametrize(
    "backend_options,account,backend,provider_name,worker_gpu",
    [
        (
            {"firewall": {"cidr": "1234.1.0.0/0"}},
            "alice",
            "vm_gcp",
            "gcp",
            None,
        ),
        (
            {"firewall": {"cidr": "1"}},
            "bobthebuilder",
            "vm",
            "aws",
            None,
        ),
    ],
)
def test_parse_backend_options_firewall_cidr_validation_raises(
    backend_options, account, backend, provider_name, worker_gpu
):
    accounts = {
        "alice": {"backend": backend},
        "hi!": {"backend": backend},
        "bobthebuilder": {"backend": backend},
    }
    with pytest.raises(CidrInvalidError):
        parse_backend_options(backend_options, account, accounts, worker_gpu)


@pytest.mark.parametrize(
    "cidr",
    [
        "1",
        "12.12.0/1",
        "0 0 0 0/0",
        "2364.985.1.0/193",
        "0.0.0.00/1234",
    ],
)
def test_validate_cidr_block_raises(cidr):
    with pytest.raises(CidrInvalidError, match="CIDR block should follow the '0.0.0.0/0' pattern."):
        validate_cidr_block(cidr)


@pytest.mark.parametrize(
    "cidr",
    [0, ["0.0.0.0/0"], {}, 1049032],
)
def test_validate_cidr_not_string(cidr):
    with pytest.raises(CidrInvalidError, match="CIDR needs to be of type string"):
        validate_cidr_block(cidr)


@pytest.mark.parametrize(
    "cidr",
    ["0.0.0.0/0", "123.456.789.101/112", "12.2.0.123/9", "98.1.0.9/984"],
)
def test_validate_cidr_block(cidr):
    is_valid = validate_cidr_block(cidr)

    assert is_valid


@pytest.mark.parametrize("ports", [[1, 234, 56, 9], [1], [22, 86787]])
def test_validate_ports(ports):
    is_valid = validate_ports(ports)
    assert is_valid


@pytest.mark.parametrize("ports", ["a", 123, 11.2])
def test_validate_ports_invalid(ports):
    with pytest.raises(PortValidationError):
        validate_ports(ports)


@pytest.mark.parametrize(
    "account",
    ["CapitalAccount", "ñjúnicode", "$?ac13", "test--account", "test-account-"],
)
def test_validate_account_errors(account):
    with pytest.raises(AccountFormatError, match="Bad account format"):
        validate_account(account)


@pytest.mark.parametrize("account", ["capitalaccount", "good-account", "should_work_too", "mix-and_match"])
def test_validate_account(account):
    result = validate_account(account)
    assert result is None


@pytest.mark.parametrize(
    "n_workers, wait_for_workers,expected",
    [
        (4, False, 0),
        (200, 0.50, 100),
        (200, 50, 50),
        (10, True, 10),
        (100, 0, 0),
        (100, None, 30),
        (1, 0, 0),
        (1, 0.3, 1),
    ],
)
def test_parse_wait_for_workers(n_workers, wait_for_workers, expected):
    to_wait = parse_wait_for_workers(n_workers, wait_for_workers)
    assert to_wait == expected


@pytest.mark.parametrize(
    "n_workers, wait_for_workers,",
    [(4, 50), (200, 3.50), (200, "hi mom!"), (10, "10"), (1, -1)],
)
def test_parse_wait_for_workers_raises(n_workers, wait_for_workers):
    with pytest.raises(ValueError, match="Received invalid value"):
        parse_wait_for_workers(n_workers, wait_for_workers)


@pytest.mark.parametrize(
    "memory, min_memory, expected",
    [
        (["12 Gib", "16 Gib"], None, {"memory__gte": 11468, "memory__lte": 18898}),
        ("12 Gib", None, {"memory__gte": 11468, "memory__lte": 14173}),
        (None, "18 Gib", {"memory__gte": 19327}),
        (24e9, None, {"memory__gte": 21360, "memory__lte": 26400}),
        (24000000000, None, {"memory__gte": 21360, "memory__lte": 26400}),
        (None, 24e9, {"memory__gte": 24000}),
        # If both `memory` annd `min_memory` is used, we use memory
        (["1Gib", "8 GiB"], "5 Gib", {"memory__gte": 956, "memory__lte": 9449}),
    ],
)
def test_parse_requested_memory(memory, min_memory, expected):
    parsed_memory = parse_requested_memory(memory, min_memory)

    assert parsed_memory == expected


@pytest.mark.parametrize(
    "bytes, expected",
    [(24000000000, 24000), (24000000000.0, 24000), (-24000000000, 24000)],
)
def test_bytes_to_mb(bytes, expected):
    assert bytes_to_mb(bytes) == expected


@pytest.mark.asyncio
async def test_handle_credentials(mocker):
    aiohttp_error = aiohttp.ContentTypeError(None, None)  # type: ignore
    bad_request = mocker.AsyncMock()
    bad_request.status = 502
    bad_request.json = mocker.AsyncMock(side_effect=aiohttp_error)

    good_request = mocker.AsyncMock()
    good_request.status = 202
    good_request.json = mocker.AsyncMock()
    good_request.json.return_value = {
        "id": 0,
        "username": "bobthebuilder",
        "email": "bob@bob.io",
        "first_name": "",
        "last_name": "",
        "membership_set": [
            {
                "is_admin": True,
                "account": {
                    "can_use_gpus": False,
                    "backend": "vm_aws",
                    "coiled_rate": 100,
                    "environment_variables": {},
                    "members_limit": 1,
                    "container_registry": {
                        "type": "ecr",
                        "has_credentials": False,
                        "public_ecr": False,
                        "region": "us-east-1",
                    },
                    "active_program": {
                        "usage_percent": 0.0,
                        "quota": 0.0,
                        "spent": 0.36,
                        "has_quota": True,
                        "start_date": "2022-02-14T13:57:41Z",
                        "end_date": "2022-03-14T13:57:41Z",
                        "tier": "enterprise",
                    },
                    "options": {"aws_region_name": "us-east-1"},
                    "id": 1,
                    "slug": "bobthebuilder",
                    "name": "bobthebuilder",
                    "limit": 150,
                    "user_limit": 150,
                    "has_quota": True,
                    "avatar_url": "",
                    "private": False,
                    "early_adopter": False,
                },
                "limit": 150,
            },
        ],
    }

    from coiled.utils import handle_credentials

    mocked_session = mocker.patch("aiohttp.ClientSession")

    mocked_session.return_value.__aenter__.return_value.request = mocker.AsyncMock(
        side_effect=[bad_request, good_request]
    )

    credentials = await handle_credentials(token="123", save=False)

    assert "bobthebuilder" in credentials
    assert "123" in credentials


@pytest.mark.asyncio
async def test_handle_credentials_no_username(mocker):
    request = mocker.AsyncMock()
    request.status = 202
    request.json = mocker.AsyncMock(return_value={})

    mocked_session = mocker.patch("aiohttp.ClientSession")

    mocked_session.return_value.__aenter__.return_value.request = mocker.AsyncMock(return_value=request)

    with pytest.raises(ValueError, match="username"):
        await handle_credentials(token="123", save=False)


@pytest.mark.asyncio
async def test_get_instance_type_from_cpu_memory_gcp(mocker):
    mocked_instance_type = mocker.patch("coiled.core.list_instance_types")
    mocked_instance_type.return_value = {
        "e2-highcpu-8": {
            "name": "e2-highcpu-8",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 8192,
            "backend_type": "vm_gcp",
        },
        "n1-standard-8": {
            "name": "n1-standard-8",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 30720,
            "backend_type": "vm_gcp",
        },
    }

    gpu_instances = get_instance_type_from_cpu_memory(cpu=8, gpus=1, backend="gcp")

    assert "n1-standard-8" in gpu_instances
    assert "e2-highcpu-8" not in gpu_instances

    no_gpu_instances = get_instance_type_from_cpu_memory(cpu=8, backend="gcp")

    assert "n1-standard-8" not in no_gpu_instances
    assert "e2-highcpu-8" in no_gpu_instances


@pytest.mark.asyncio
async def test_get_instance_type_from_cpu_memory_aws(mocker):
    mocked_instance_type = mocker.patch("coiled.core.list_instance_types")
    mocked_instance_type.return_value = {
        "g4dn.2xlarge": {
            "name": "g4dn.2xlarge",
            "cores": 8,
            "gpus": 1,
            "gpu_name": "T4",
            "memory": 32768,
            "backend_type": "vm_aws",
        },
        "t3.2xlarge": {
            "name": "t3a.2xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 32768,
            "backend_type": "vm_aws",
        },
    }

    gpu_instances = get_instance_type_from_cpu_memory(cpu=8, gpus=1, backend="aws")

    assert "g4dn.2xlarge" in gpu_instances
    assert "t3a.2xlarge" not in gpu_instances

    no_gpu_instances = get_instance_type_from_cpu_memory(cpu=8, backend="aws")

    assert "g4dn.2xlarge" not in no_gpu_instances
    assert "t3.2xlarge" in no_gpu_instances


@pytest.mark.asyncio
async def test_get_instance_type_from_cpu_memory_aws_filter(mocker):
    mocked_instance_type = mocker.patch("coiled.core.list_instance_types")
    mocked_instance_type.return_value = {
        "g4dn.2xlarge": {
            "name": "g4dn.2xlarge",
            "cores": 8,
            "gpus": 1,
            "gpu_name": "T4",
            "memory": 32768,
            "backend_type": "vm_aws",
        },
        "t3.2xlarge": {
            "name": "t3a.2xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 32768,
            "backend_type": "vm_aws",
        },
        "t3a.2xlarge": {
            "name": "t3a.xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 32768,
            "backend_type": "vm_aws",
        },
    }

    instances = get_instance_type_from_cpu_memory(cpu=8, backend="aws")

    assert "t3.2xlarge" in instances
    assert "t3a.2xlarge" not in instances


@pytest.mark.asyncio
async def test_get_instance_type_from_cpu_memory_aws_unbalanced_filter(mocker):
    mocked_instance_type = mocker.patch("coiled.core.list_instance_types")
    mocked_instance_type.return_value = {
        "g4dn.2xlarge": {
            "name": "g4dn.2xlarge",
            "cores": 8,
            "gpus": 1,
            "gpu_name": "T4",
            "memory": 32768,
            "backend_type": "vm_aws",
        },
        "t3.2xlarge": {
            "name": "t3a.2xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 32768,
            "backend_type": "vm_aws",
        },
        "c6i.2xlarge": {
            "name": "c6i.2xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 16384,
            "backend_type": "vm_aws",
        },
        "c5.2xlarge": {
            "name": "c5.2xlarge",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 16384,
            "backend_type": "vm_aws",
        },
    }

    # If we don't pass memory, we shouldn't get any unbalanced instances
    instances = get_instance_type_from_cpu_memory(cpu=8, backend="aws")

    assert "t3.2xlarge" in instances
    assert "c6i.2xlarge" not in instances
    assert "c5.2xlarge" not in instances

    # If we *do* pass memory, we should get unbalanced and balanced instances
    instances = get_instance_type_from_cpu_memory(cpu=8, memory="16gb", backend="aws")

    assert "t3.2xlarge" in instances
    assert "c6i.2xlarge" in instances
    assert "c5.2xlarge" in instances



import asyncio
import logging
import pkgutil
import sys
from logging import basicConfig
from pathlib import Path

import click
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from coiled.scan import scan_prefix


@click.group()
def package_sync():
    basicConfig(level=logging.INFO)


@package_sync.command()
def scan():
    result = asyncio.run(scan_prefix(Path(sys.prefix)))
    table = Table(title="Packages")
    table.add_column("Package Name", style="cyan", no_wrap=True)
    table.add_column("Conda Name", style="cyan", no_wrap=True)
    table.add_column("Version", style="magenta")
    table.add_column("Source", style="magenta")
    table.add_column("Wheel target", style="green", overflow="fold")
    table.add_column("Path", overflow="fold")
    table.add_column("Can build wheel", style="green")
    rows = []
    for pkg in result:
        rows.append(
            (
                pkg["name"],
                pkg["conda_name"],
                pkg["version"],
                pkg["source"],
                str(pkg["wheel_target"] or ""),
                str(pkg["path"]),
                "true",
            )
        )

    for _, name, ispkg in pkgutil.iter_modules(path=["."]):
        will_sync = False
        if ispkg:
            for pkg in result:
                if pkg["path"]:
                    if (pkg["path"] / name).resolve() == (Path(".") / name).resolve():  # type: ignore
                        will_sync = True
        if not will_sync:
            rows.append(
                (
                    name,
                    "",
                    "cwd",
                    "",
                    str(Path(".") / name) if ispkg else str((Path(".") / name).with_suffix(".py")),
                    "[red]false[/red]",
                )
            )
    for row in sorted(rows, key=lambda x: (x[5], x[0].lower())):
        table.add_row(*row)
    console = Console()
    console.print(table)
    console.print(
        Panel(
            "[yellow]Warning: You have importable code in your path"
            " which won't be available on your cluster because we're"
            " unable to build wheel.[/yellow]"
        )
    )


@package_sync.command()
def debug():
    table = Table(title="Debug")
    table.add_column("Path", no_wrap=True, overflow="fold")
    rows = []
    for path in sys.path:
        p = Path(path)
        if p.is_dir():
            for file in p.iterdir():
                rows.append(str(file))
        else:
            rows.append(str(p))
    rows = sorted(rows)
    for row in rows:
        table.add_row(row)
    console = Console()
    console.print(table)



import asyncio
import logging
import platform
import re
import subprocess
import typing
from collections import defaultdict
from contextlib import asynccontextmanager
from logging import getLogger
from pathlib import Path
from sys import executable
from tempfile import TemporaryDirectory, gettempdir
from typing import Dict, List, Optional, Set, TextIO, Union
from urllib.parse import urlparse

from filelock import BaseFileLock, FileLock
from packaging.utils import parse_wheel_filename
from rich.progress import Progress

from coiled._beta.core import CloudBeta
from coiled._beta.widgets.util import simple_progress
from coiled.scan import PackageInfo, scan_prefix
from coiled.types import PackageLevelEnum, ResolvedPackageInfo
from coiled.utils import validate_wheel

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()
ANY_AVAILABLE = "ANY-AVAILABLE"


async def create_subprocess_shell(
    cmd: str,
    stdout: Union[TextIO, int, None] = None,
    stderr: Union[TextIO, int, None] = None,
) -> subprocess.CompletedProcess:
    # asyncio.create_subprocess_shell is
    # not supported on windows
    # and broken on py3.7
    loop = asyncio.get_running_loop()
    result = loop.run_in_executor(None, lambda: subprocess.run(cmd, shell=True, stdout=stdout, stderr=stderr))
    return await result


class PackageBuildError(Exception):
    pass


WHEEL_BUILD_LOCKS: Dict[str, BaseFileLock] = {}


async def default_python() -> PackageInfo:
    python_version = platform.python_version()
    return {
        "name": "python",
        "path": None,
        "source": "conda",
        "channel_url": ANY_AVAILABLE,
        "channel": ANY_AVAILABLE,
        "subdir": "linux-64",
        "conda_name": "python",
        "version": python_version,
        "wheel_target": None,
    }


@asynccontextmanager
async def async_lock(lock: BaseFileLock):
    loop = asyncio.get_event_loop()
    ctx = await loop.run_in_executor(None, lock.acquire)
    with ctx:
        yield


async def create_wheel(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    lock = WHEEL_BUILD_LOCKS.setdefault(pkg_name, FileLock(Path(gettempdir()) / (pkg_name + version + ".lock")))
    async with async_lock(lock):
        tmpdir = TemporaryDirectory()
        outdir = Path(tmpdir.name) / Path(pkg_name)
        logger.info(f"Attempting to create a wheel for {pkg_name} @ {src}")
        # must use executable to avoid using some other random python
        p = await create_subprocess_shell(
            cmd=f"{executable} -m pip wheel --wheel-dir {outdir} --no-deps --use-pep517 --no-cache-dir {src}",
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
        if p.returncode:
            print(f"---Wheel Build Log for {pkg_name}---\n{p.stdout.decode()}")
            return {
                "name": pkg_name,
                "source": "pip",
                "channel": None,
                "conda_name": None,
                "client_version": version,
                "specifier": "",
                "include": False,
                "error": (
                    "Failed to build a wheel for the"
                    " package, will not be included in environment, check stdout for the build log"
                ),
                "note": None,
                "sdist": None,
                "md5": None,
            }
        wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5 = await validate_wheel(wheel_fn)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel contains no python files!",
        "note": f"Wheel built from {src}",
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


async def create_wheel_from_egg(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    tmpdir = TemporaryDirectory()
    outdir = Path(tmpdir.name) / Path(pkg_name)
    outdir.mkdir(parents=True)
    logger.info(f"Attempting to create a wheel for {pkg_name} in directory {src}")
    # must use executable to avoid using some other random python
    p = await create_subprocess_shell(
        cmd=f"{executable} -m wheel convert --dest-dir {outdir} {src}",
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )
    if p.returncode:
        print(f"---Egg to wheel conversion Log for {pkg_name}---\n{p.stdout.decode()}")
        return {
            "name": pkg_name,
            "source": "pip",
            "channel": None,
            "conda_name": None,
            "client_version": version,
            "specifier": "",
            "include": False,
            "error": (
                "Failed to convert the package egg to a wheel"
                ", will not be included in environment, check stdout for egg conversion log"
            ),
            "note": None,
            "sdist": None,
            "md5": None,
        }
    wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
    has_python, md5 = await validate_wheel(Path(wheel_fn))
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": version,
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel has no python files!",
        "note": "Wheel built from local egg",
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


async def approximate_packages(
    cloud: CloudBeta,
    packages: List[PackageInfo],
    priorities: Dict[str, PackageLevelEnum],
    progress: Optional[Progress] = None,
    strict: bool = False,
) -> typing.List[ResolvedPackageInfo]:
    user_conda_installed_python = next((p for p in packages if p["name"] == "python"), None)
    user_conda_installed_pip = next(
        (i for i, p in enumerate(packages) if p["name"] == "pip" and p["source"] == "conda"),
        None,
    )
    if not user_conda_installed_pip:
        # This means pip was installed by pip, or the system
        # package manager
        # Insert a conda version of pip to be installed first, it will
        # then be used to install the users version of pip
        pip = next(
            (p for p in packages if p["name"] == "pip" and p["source"] == "pip"),
            None,
        )
        if not pip:
            # insert a modern version and hope it does not introduce conflicts
            packages.append(
                {
                    "name": "pip",
                    "path": None,
                    "source": "conda",
                    "channel_url": "https://conda.anaconda.org/conda-forge/",
                    "channel": "conda-forge",
                    "subdir": "noarch",
                    "conda_name": "pip",
                    "version": "22.3.1",
                    "wheel_target": None,
                }
            )
        else:
            # insert the users pip version and hope it exists on conda-forge
            packages.append(
                {
                    "name": "pip",
                    "path": None,
                    "source": "conda",
                    "channel_url": "https://conda.anaconda.org/conda-forge/",
                    "channel": "conda-forge",
                    "subdir": "noarch",
                    "conda_name": "pip",
                    "version": pip["version"],
                    "wheel_target": None,
                }
            )
    coiled_selected_python = None
    if not user_conda_installed_python:
        # insert a special python package
        # that the backend will pick a channel for
        coiled_selected_python = await default_python()
        packages.append(coiled_selected_python)
    with simple_progress("Validating environment", progress=progress):
        results = await cloud._approximate_packages(
            packages=[
                {
                    "name": pkg["name"],
                    "priority_override": 100 if (strict or pkg["wheel_target"]) else priorities.get(pkg["name"]),
                    "python_major_version": PYTHON_VERSION[0],
                    "python_minor_version": PYTHON_VERSION[1],
                    "python_patch_version": PYTHON_VERSION[2],
                    "source": pkg["source"],
                    "channel_url": pkg["channel_url"],
                    "channel": pkg["channel"],
                    "subdir": pkg["subdir"],
                    "conda_name": pkg["conda_name"],
                    "version": pkg["version"],
                    "wheel_target": pkg["wheel_target"],
                }
                for pkg in packages
            ]
        )
    result_map = {r["name"]: r for r in results}
    finalized_packages: typing.List[ResolvedPackageInfo] = []

    if not user_conda_installed_python and coiled_selected_python:
        # user has no python version installed by conda
        # we should have a result of asking the backend to
        # pick conda channel that has the users python version
        if result_map["python"]["error"]:
            finalized_packages.append(
                {
                    "name": "python",
                    "source": "conda",
                    "channel": None,
                    "conda_name": "python",
                    "client_version": coiled_selected_python["version"],
                    "specifier": result_map["python"]["specifier"] or "",
                    "include": result_map["python"]["include"],
                    "note": None,
                    "error": result_map["python"]["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
        else:
            assert result_map["python"]["note"]
            channel_url, channel = result_map["python"]["note"].split(",")
            finalized_packages.append(
                {
                    "name": "python",
                    "source": "conda",
                    "channel": channel,
                    "conda_name": "python",
                    "client_version": coiled_selected_python["version"],
                    "specifier": result_map["python"]["specifier"] or "",
                    "include": result_map["python"]["include"],
                    "note": None,
                    "error": result_map["python"]["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
        # we can pull our special python package out of the list
        # now
        packages.remove(coiled_selected_python)
    for pkg in packages:

        if pkg["wheel_target"] and result_map[pkg["name"]]["include"]:
            p = urlparse(pkg["wheel_target"])
            if p.scheme == "" and not Path(pkg["wheel_target"]).exists():
                # lack of scheme means a local file
                # sometimes the wheel target taken from
                # direct_url.json references a file that does not exist
                # skip over trying to sync that and treat it like a normal package
                # this can happen in conda/system packages
                # where the package metadata was generated on a build system and not locally
                finalized_packages.append(
                    {
                        "name": pkg["name"],
                        "source": pkg["source"],
                        "channel": pkg["channel"],
                        "conda_name": pkg["conda_name"],
                        "client_version": pkg["version"],
                        "specifier": result_map[pkg["name"]]["specifier"] or "",
                        "include": result_map[pkg["name"]]["include"],
                        "note": result_map[pkg["name"]]["note"],
                        "error": result_map[pkg["name"]]["error"],
                        "sdist": None,
                        "md5": None,
                    }
                )
            elif pkg["wheel_target"].endswith(".egg"):
                with simple_progress(f'Creating wheel from egg for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel_from_egg(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
            else:
                with simple_progress(f'Creating wheel for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
        else:
            finalized_packages.append(
                {
                    "name": pkg["name"],
                    "source": pkg["source"],
                    "channel": pkg["channel"],
                    "conda_name": pkg["conda_name"],
                    "client_version": pkg["version"],
                    "specifier": result_map[pkg["name"]]["specifier"] or "",
                    "include": result_map[pkg["name"]]["include"],
                    "note": result_map[pkg["name"]]["note"],
                    "error": result_map[pkg["name"]]["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
    return finalized_packages


pip_bad_req_regex = (
    r"(?P<package>.+) (?P<version>.+) has requirement "
    r"(?P<requirement>.+), but you have (?P<requirement2>.+) (?P<reqversion>.+)."
)


async def check_pip_happy(progress: Optional[Progress] = None) -> Dict[str, List[str]]:
    with simple_progress("Running pip check", progress=progress):
        proc = await create_subprocess_shell(
            f"{executable} -m pip check", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        faulty_packages = defaultdict(list)
        if proc.returncode:
            output = proc.stdout.decode()
            bad_reqs = re.finditer(pip_bad_req_regex, output)
            for bad_req in bad_reqs:
                groups = bad_req.groupdict()
                span = bad_req.span()
                warning = output[span[0] : span[1]]
                logger.warning(warning)
                faulty_packages[groups["package"]].append(warning)
        return faulty_packages


async def create_environment_approximation(
    cloud: CloudBeta,
    priorities: Dict[str, PackageLevelEnum],
    only: Optional[Set[str]] = None,
    strict: bool = False,
    progress: Optional[Progress] = None,
) -> typing.List[ResolvedPackageInfo]:
    packages = await scan_prefix(progress=progress)
    bad_packages = await check_pip_happy(progress)
    if bad_packages:
        packages = [pkg for pkg in packages if pkg["name"] not in bad_packages]
    if only:
        packages = filter(lambda pkg: pkg["name"] in only, packages)
    # TODO: private conda channels
    result = await approximate_packages(
        cloud=cloud, packages=[pkg for pkg in packages], priorities=priorities, strict=strict, progress=progress
    )
    for bad_package, error_list in bad_packages.items():
        errors = "\n".join(error_list)
        result.append(
            {
                "name": bad_package,
                "error": f"Pip check had the following issues that need resolving: \n{errors}",
                "source": "pip",
                "channel": None,
                "client_version": "",
                "conda_name": None,
                "include": False,
                "md5": None,
                "note": None,
                "sdist": None,
                "specifier": "",
            }
        )
    return result


if __name__ == "__main__":
    from logging import basicConfig

    basicConfig(level=logging.INFO)

    from rich.console import Console
    from rich.table import Table

    async def run():
        async with CloudBeta(asynchronous=True) as cloud:
            return await create_environment_approximation(
                cloud=cloud,
                priorities={
                    "dask": PackageLevelEnum.CRITICAL,
                    "twisted": PackageLevelEnum.IGNORE,
                    "graphviz": PackageLevelEnum.OK,
                    "icu": PackageLevelEnum.OK,
                },
            )

    result = asyncio.run(run())

    table = Table(title="Packages")
    keys = ("name", "source", "include", "client_version", "specifier", "error", "note")

    for key in keys:
        table.add_column(key)

    for pkg in result:
        row_values = [str(pkg.get(key, "")) for key in keys]
        table.add_row(*row_values)
    console = Console()
    console.print(table)
    console.print(table)



#!/usr/bin/env python

from os.path import exists

import versioneer
from pkg_resources import parse_requirements
from setuptools import find_packages, setup

setup(
    name="coiled",
    url="https://coiled.io",
    maintainer="Coiled",
    maintainer_email="info@coiled.io",
    version=versioneer.get_version(),  # type: ignore
    cmdclass=versioneer.get_cmdclass(),  # type: ignore
    description="",
    packages=find_packages(),
    include_package_data=True,
    long_description=(open("README.md").read() if exists("README.md") else ""),
    long_description_content_type="text/markdown",
    zip_safe=False,
    install_requires=[str(requirement) for requirement in parse_requirements(open("requirements.txt").read())],
    entry_points={"console_scripts": ["coiled=coiled.cli.core:cli"]},
    python_requires=">=3.7",
)



import shutil
import subprocess
import sys
from datetime import datetime
from typing import Optional

import click
from rich import print

import coiled

from ..utils import CONTEXT_SETTINGS


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("cluster")
@click.option(
    "--account",
    default=None,
    help="Coiled account (uses default account if not specified)",
)
@click.option(
    "--scheduler",
    default=False,
    is_flag=True,
    help="Get scheduler logs",
)
@click.option(
    "--workers",
    help="Get worker logs ('any', 'all', or comma-delimited list of names, states, or internal IP addresses)",
)
@click.option(
    "--follow",
    default=False,
    is_flag=True,
    help="Passed directly to `aws logs tail`, see aws cli docs for details.",
)
@click.option(
    "--filter",
    default=None,
    help="Passed directly to `aws logs tail`, see aws cli docs for details.",
)
@click.option(
    "--since",
    default=None,
    help="For follow, uses `aws logs tail` default (10m), otherwise defaults to start time of cluster.",
)
@click.option(
    "--format",
    default=None,
    help="Passed directly to `aws logs tail`, see aws cli docs for details.",
)
@click.option(
    "--profile",
    default=None,
    help="Passed directly to `aws logs tail`, see aws cli docs for details.",
)
def logs(
    cluster: str,
    account: Optional[str],
    scheduler: bool,
    workers: Optional[str],
    follow: bool,
    filter: Optional[str],
    since: Optional[str],
    format: Optional[str],
    profile: Optional[str],
):
    do_logs(cluster, account, scheduler, workers, follow, filter, since, format, profile)


def do_logs(
    cluster: str,
    account: Optional[str],
    scheduler: bool,
    workers: Optional[str],
    follow: bool,
    filter: Optional[str],
    since: Optional[str],
    format: Optional[str],
    profile: Optional[str],
    capture: bool = False,
):
    if not shutil.which("aws"):
        raise click.ClickException("`coiled cluster logs` relies on AWS CLI.")

    with coiled.Cloud(account=account) as cloud:
        if cluster.isnumeric():
            cluster_id = int(cluster)
        else:
            try:
                clusters = cloud.get_clusters_by_name(name=cluster)
                if clusters:
                    recent_cluster = clusters[-1]
                else:
                    raise click.ClickException(f"Unable to find cluster with name '{cluster}'")

                if follow and recent_cluster["current_state"]["state"] in (
                    "stopped",
                    "error",
                ):
                    follow = False
                    print(
                        f"[red]Cluster state is {recent_cluster['current_state']['state']} so not following.[/red]",
                        file=sys.stderr,
                    )

                cluster_id = recent_cluster["id"]

            except coiled.errors.DoesNotExist:
                cluster_id = None

        if not cluster_id:
            raise click.ClickException(f"Unable to find cluster `{cluster}`")

        cluster_info = cloud.cluster_details(cluster_id)

        # for tailing, use aws default (10m) but allow "start"; for non-tailing, default to start of cluster
        if (since is None and not follow) or since == "start":
            cluster_start_ts = datetime.fromisoformat(cluster_info["created"]).timestamp()
            since = str(int(cluster_start_ts))

        if workers:
            worker_attrs_to_match = workers.split(",")

            def filter_worker(idx, worker):
                if workers == "all":
                    return True
                elif workers == "any":
                    if idx == 0:
                        return True
                else:
                    if worker.get("name") and worker["name"] in worker_attrs_to_match:
                        return True
                    elif (
                        worker.get("instance", {}).get("private_ip_address")
                        and worker["instance"]["private_ip_address"] in worker_attrs_to_match
                    ):
                        return True
                    elif (
                        worker.get("current_state", {}).get("state")
                        and worker["current_state"]["state"] in worker_attrs_to_match
                    ):
                        return True

                return False

            worker_names = [
                worker["name"] for i, worker in enumerate(cluster_info["workers"]) if filter_worker(i, worker)
            ]
        else:
            worker_names = []

        log_info = cloud.get_cluster_log_info(cluster_id)

        if log_info["type"] == "vm_aws":
            group_name = log_info["log_group"]
            region = log_info["region"]

            stream_names = []

            if not group_name:
                raise click.ClickException("Unable to find CloudWatch Log Group for this cluster.")

            if scheduler:
                stream_names.append(log_info["scheduler_stream"])

            if workers:
                worker_stream_name = [log_info["worker_streams"][worker_name] for worker_name in worker_names]
                stream_names.extend(worker_stream_name)

            if not stream_names:
                raise click.ClickException(
                    "No CloudWatch Log Streams to show, use `--scheduler` or `--workers` to select streams."
                )  # FIXME better error

            streams = " ".join(stream_names)

            command = f"aws logs tail {group_name} --region {region} --log-stream-names {streams}"
            if follow:
                command = f"{command} --follow"
            if filter:
                command = f"{command} --filter {filter}"
            if since:
                command = f"{command} --since {since}"
            if format:
                command = f"{command} --format {format}"
            if profile:
                command = f"{command} --profile {profile}"

            return subprocess.run(command.split(" "), capture_output=capture)

            # TODO `aws logs tail` gives ResourceNotFoundException error ("The specified log stream does not exist.")
            #   when *any* of the log streams doesn't exist. We should query to see which exist before trying to tail.

        else:
            raise click.ClickException(f"Cluster backend type is {log_info['type']}, only AWS is currently supported.")



import dask
from prefect import flow, task
from prefect_dask import DaskTaskRunner, get_dask_client


@task
def compute_task():
    with get_dask_client() as client:  # noqa
        df = dask.datasets.timeseries("2000", "2001", partition_freq="4w")
        summary_df = df.describe().compute()
    return summary_df


@flow(
    task_runner=DaskTaskRunner(
        cluster_class="coiled.Cluster",
        cluster_kwargs={
            "name": "",
            "n_workers": 4,
            "software": "prefect2-example",
        },
    )
)
def dask_flow():
    prefect_future = compute_task.submit()
    return prefect_future.result()


if __name__ == "__main__":
    dask_flow()



import sys
from distutils.version import LooseVersion

import distributed

from ._version import get_versions

COILED_VERSION = get_versions()["version"]

PY_VERSION = LooseVersion(".".join(map(str, sys.version_info[:3])))

DISTRIBUTED_VERSION = LooseVersion(distributed.__version__)



from __future__ import annotations, with_statement

from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime, timezone
from logging import getLogger
from typing import Any, Callable, Coroutine, Dict, TypeVar

import aiohttp
from typing_extensions import ParamSpec

from .utils import get_datadog_trace_link, random_str

try:
    from ddtrace import tracer
    from ddtrace.propagation.http import HTTPPropagator

    DDTRACE = True

except ImportError:
    DDTRACE = False

logger = getLogger(__name__)

# Sessions last the entire duration of the python process
COILED_SESSION_ID = "coiled-session-" + random_str()
logger.debug(f"Coiled Session  ID : {COILED_SESSION_ID}")

# Operations are transient and more granular
# Note: we don't type the actual RHS value due to a bug in some versions of Python
# 3.7 and 3.8: https://bugs.python.org/issue38979
COILED_OP_CONTEXT: ContextVar[str] = ContextVar("coiled-operation-context", default="")
COILED_OP_NAME: ContextVar[str] = ContextVar("coiled-operation-name", default="")

ContextReturnT = TypeVar("ContextReturnT")

TRACE_CONFIG = aiohttp.TraceConfig()


@contextmanager
def operation_context(name: str):
    c_id = COILED_OP_CONTEXT.get()
    if c_id:
        # already in a coiled op context, don't create a new one
        yield c_id
    else:
        # create a new coiled context
        c_id = random_str()
        COILED_OP_CONTEXT.set(c_id)
        COILED_OP_NAME.set(name)

        logger.debug(f"Entering {name}-{c_id}")
        start = datetime.now(tz=timezone.utc)
        yield c_id
        trace_url = get_datadog_trace_link(
            start=start,
            end=datetime.now(tz=timezone.utc),
            **{"coiled-operation-id": c_id},
        )
        logger.debug(f"Exiting {name}-{c_id} - DD URL: {trace_url}")
        COILED_OP_CONTEXT.set("")
        COILED_OP_NAME.set("")


TRACK_PARAMS = ParamSpec("TRACK_PARAMS")


def track_context(
    func: Callable[TRACK_PARAMS, Coroutine[Any, Any, ContextReturnT]]
) -> Callable[TRACK_PARAMS, Coroutine[Any, Any, ContextReturnT]]:
    async def wrapper(*args: TRACK_PARAMS.args, **kwargs: TRACK_PARAMS.kwargs):
        if DDTRACE and tracer.current_span():
            with tracer.trace(name=f"{func.__module__}.{func.__qualname__}"), operation_context(
                name=f"{func.__module__}.{func.__qualname__}"
            ):
                return await func(*args, **kwargs)
        else:
            with operation_context(name=f"{func.__module__}.{func.__qualname__}"):
                return await func(*args, **kwargs)

    return wrapper


def inject_tracing(headers: Dict[str, str]):
    headers.update(create_trace_data())


def create_trace_data() -> Dict[str, str]:
    trace_data: Dict[str, str] = {
        "coiled-session-id": COILED_SESSION_ID,
        "coiled-request-id": random_str(),
    }
    if DDTRACE:
        span = tracer.current_span()
        if span:
            HTTPPropagator.inject(span_context=span.context, headers=trace_data)
    op_id = COILED_OP_CONTEXT.get()
    if op_id:
        trace_data["coiled-operation-id"] = op_id
    op_name = COILED_OP_CONTEXT.get()
    if op_name:
        trace_data["coiled-operation-func"] = op_name
    return trace_data


async def on_request_start(session, trace_config_ctx, params):
    inject_tracing(headers=params.headers)


TRACE_CONFIG.on_request_start.append(on_request_start)



import click

import coiled

from .utils import CONTEXT_SETTINGS


@click.command(
    context_settings=CONTEXT_SETTINGS,
    help="CLI to hit endpoints using Coiled account authentication (mostly for internal use)",
)
@click.argument("url")
@click.option("-X", "--request", default="GET")
@click.option("-d", "--data", multiple=True)
def curl(url: str, request, data):
    all_data = "&".join(data) if data else None
    with coiled.Cloud() as cloud:
        if "http" not in url:
            url = f"{cloud.server}{url}"

        response = sync_request(cloud, url, method=request, data=all_data)
    print(response)


def sync_request(cloud, url, method, data):
    response = cloud._sync(
        cloud._do_request,
        method=method,
        url=url,
        data=data,
    )
    if response.status >= 400:
        print(f"{url} returned {response.status}")

    async def get_text(r):
        return await r.text()

    return cloud._sync(
        get_text,
        response,
    )
